import os
import pytest
from glob import glob
from sqlalchemy import Column
from sqlalchemy.orm import Session
from sqlalchemy.exc import NoSuchTableError
from unittest import TestCase
from .seed_db import create_table
from lazy_alchemy import get_lazy_class, CustomTable


class TestLazyOrmSuite(TestCase):
    def setUp(self) -> None:
        self.engine = create_table()
        self.session = Session(self.engine)
        self.lazy_class = get_lazy_class(self.engine)

    def test_lazy_db_connection(self):
        table = self.lazy_class.user
        self.assertIsInstance(table, CustomTable)
        self.assertIsInstance(table.username, Column)
        self.assertIsInstance(table.age, Column)
        self.assertEqual(table.name, "user")
        self.assertEqual(len(table.constraints), 1)
        self.assertTrue(
            any(i for i in table.indexes if i.name == "idx_user_username"))

    def test_insert_operation_on_dynamic_class(self):
        user: CustomTable = self.lazy_class.user
        self.assertTrue(user)
        insert_statement = user.insert().values(username="fake_user", age=21)
        self.session.execute(insert_statement)
        self.session.commit()
        obj = self.session.query(user).first()
        self.assertEqual(obj.username, "fake_user")
        self.assertEqual(obj.age, 21)

    def test_invalid_table_and_column(self):
        with pytest.raises(NoSuchTableError):
            self.lazy_class.foo_bar
        with pytest.raises(AttributeError):
            self.lazy_class.user.abc

    def test_overriding_table(self):
        _ = self.lazy_class.user
        self.lazy_class.user = CustomTable("userabc", self.lazy_class.metadata)
        self.assertNotEqual(_, self.lazy_class.user)
        self.assertEqual(self.lazy_class.user.name, "userabc")

    def tearDown(self) -> None:
        self.session.close()
        print(f"cleaning files {glob('*.db')}")
        os.unlink("test.db")



from curses import echo
from sqlalchemy import (create_engine, Column, String, Integer, Index, engine)
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class User(Base):
    __tablename__ = "user"
    username = Column("username", String(20), primary_key=True)
    age = Column("age", Integer)
    __table_args__ = (
        Index("idx_user_username", "username"),
    )


def create_table():
    e: engine.Engine = create_engine(
        "sqlite:///test.db", echo=True, future=True)
    with e.begin() as c:
        Base.metadata.create_all(bind=c)
    return e



from pathlib import Path

from lazy_alchemy.lazy_alchemy import get_lazy_class, CustomTable

here = Path(__file__).resolve().parent
version = (here / "VERSION").read_text(encoding="utf-8")

__version__ = version

__all__ = ("get_lazy_class", "CustomTable", "version")



from sqlalchemy import Table, MetaData
from sqlalchemy.engine import Engine


class CustomTable(Table):
    def __getattr__(self, attr):
        return getattr(self.c, attr)

    def __bool__(self) -> None:
        return self is not None


class LazyDBProp(object):
    """This descriptor returns sqlalchemy
    Table class which can be used to query
    table from the schema
    """

    def __init__(self) -> None:
        self._table = None
        self._name = None

    def __set_name__(self, _, name):
        self._name = name

    def __set__(self, instance, value):
        if isinstance(value, (CustomTable, Table)):
            self._table = value

    def __get__(self, instance, _):
        if self._table is None:
            self._table = CustomTable(
                self._name, instance.metadata, autoload=True)
        return self._table


def get_lazy_class(engine: Engine) -> object:
    """
    Function to create Lazy class for pulling table object
    using SQLalchemy metadata
    """

    def __init__(self, engine: Engine):
        self.metadata = MetaData(engine)
        self.engine = engine

    def __getattr__(self, attr):
        if attr not in self.__dict__:
            obj = self.__patch(attr)
        return obj.__get__(self, type(self))

    def __patch(self, attribute):
        obj = LazyDBProp()
        obj.__set_name__(self, attribute)
        setattr(type(self), attribute, obj)
        return obj

    # naming classes uniquely for different schema's
    # to avoid cross referencing
    LazyClass = type(f"LazyClass_{engine.url.database}", (), {})
    LazyClass.__init__ = __init__
    LazyClass.__getattr__ = __getattr__
    LazyClass.__patch = __patch
    return LazyClass(engine)



from pathlib import Path

from setuptools import setup, find_packages

from lazy_alchemy import version

here = Path(__file__).resolve().parent
README = (here / "README.md").read_text(encoding="utf-8")
PACKAGE_NAME = "lazy_alchemy"

setup(
    name=PACKAGE_NAME,
    version=version,
    license="MIT",
    packages=find_packages(exclude=["test"]),
    description="Lazy-Alchemy is a Python package that loads the DB models lazily.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Satyam Soni",
    author_email="satyamsoni@hotmail.co.uk",
    url="https://github.com/satyamsoni2211/lazy_alchemy",
    keywords=["sqlalchemy", "alchemy", "mysql", "postgres",
              "mssql", "sql", "sqlite", "lazy", "performance",
              "orm", "mapper", "performance", "database", "lazy",
              "relational", "classes", "oops", "metaclass"],
    install_requires=[
        "sqlalchemy",
    ],
    classifiers=[
        # See https://pypi.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    platforms=["any"],
    python_requires=">=3.6",
)
